# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import base64
import collections
import io
import numbers
import traceback

import numpy as np
import six

from . import viewer_state
from .json_wrappers import (JsonObjectWrapper, array_wrapper, optional, text_type, typed_list,
                            typed_set, typed_string_map, wrapped_property)

_uint64_keys = frozenset(['t', 'v'])
_map_entry_keys = frozenset(['key', 'value'])


class SegmentIdMapEntry(collections.namedtuple('SegmentIdMapEntry', ['key', 'value', 'label'])):
    def __new__(cls, key, value=None, label=None):
        return super(SegmentIdMapEntry, cls).__new__(cls, key, value, label)


def layer_selected_value(x):
    if isinstance(x, numbers.Number):
        return x
    if isinstance(x, six.string_types):
        return int(x)
    if isinstance(x, dict):
        value = x.get('value')
        if value is not None:
            value = int(value)
        return SegmentIdMapEntry(int(x['key']), value, x.get('label'))
    return None


class LayerSelectionState(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    local_position = wrapped_property('localPosition', optional(array_wrapper(np.float32)))
    value = wrapped_property('value', optional(layer_selected_value))


LayerSelectedValues = typed_string_map(LayerSelectionState)


class ScreenshotReply(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property('id', text_type)
    image = wrapped_property('image', base64.b64decode)
    width = wrapped_property('width', int)
    height = wrapped_property('height', int)
    image_type = imageType = wrapped_property('imageType', text_type)
    depth_data = depthData = wrapped_property('depthData', optional(base64.b64decode))

    @property
    def image_pixels(self):
        """Returns the screenshot image as a numpy array of pixel values."""
        import PIL
        return np.asarray(PIL.Image.open(io.BytesIO(self.image)))

    @property
    def depth_array(self):
        """Returns the depth data as a numpy float32 array."""
        depth_data = self.depth_data
        if depth_data is None: return None
        return np.frombuffer(depth_data, dtype='<f4').reshape((self.height, self.width))


class AggregateChunkSourceStatistics(JsonObjectWrapper):
    __slots__ = ()
    visible_chunks_total = visibleChunksTotal = wrapped_property('visibleChunksTotal', int)
    visible_chunks_downloading = visibleChunksDownloading = wrapped_property(
        'visibleChunksDownloading', int)
    visible_chunks_system_memory = visibleChunksSystemMemory = wrapped_property(
        'visibleChunksSystemMemory', int)
    visible_chunks_gpu_memory = visibleChunksGpuMemory = wrapped_property(
        'visibleChunksGpuMemory', int)
    visible_gpu_memory = visibleGpuMemory = wrapped_property('visibleGpuMemory', float)
    download_latency = downloadLatency = wrapped_property('downloadLatency', float)


class ChunkSourceStatistics(JsonObjectWrapper):
    __slots__ = ()
    distinct_id = distinctId = wrapped_property('distinctId', text_type)


class ScreenshotStatistics(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property('id', text_type)

    chunk_sources = chunkSources = wrapped_property('chunkSources',
                                                    typed_list(ChunkSourceStatistics))
    total = wrapped_property('total', AggregateChunkSourceStatistics)


class ActionState(JsonObjectWrapper):
    __slots__ = ()
    viewer_state = viewerState = wrapped_property('viewerState', viewer_state.ViewerState)
    selected_values = selectedValues = wrapped_property('selectedValues', LayerSelectedValues)
    mouse_position = mouse_voxel_coordinates = mouseVoxelCoordinates = wrapped_property(
        'mousePosition', optional(array_wrapper(np.float32)))
    screenshot = wrapped_property('screenshot', optional(ScreenshotReply))
    screenshot_statistics = screenshotStatistics = wrapped_property('screenshotStatistics',
                                                                    optional(ScreenshotStatistics))


class Actions(object):
    def __init__(self, set_config):
        self._action_handlers = dict()
        self._set_config = set_config

    def add(self, name, handler):
        self._action_handlers.setdefault(name, set()).add(handler)
        self._update_config()

    def clear(self):
        screenshot_handler = self._action_handlers.get('screenshot')
        self._action_handlers.clear()
        if screenshot_handler is not None:
            self._action_handlers['screenshot'] = screenshot_handler
        self._update_config()

    def remove(self, name, handler):
        handlers = self._action_handlers.get(name)
        if handlers is None:
            return
        handlers.remove(handler)
        if not handlers:
            del self._action_handlers[name]
        self._update_config()

    def _update_config(self):
        self._set_config(six.viewkeys(self._action_handlers))

    def invoke(self, name, state):
        state = ActionState(state)
        handlers = self._action_handlers.get(name)
        if handlers is not None:
            for handler in handlers:
                try:
                    handler(state)
                except:
                    traceback.print_exc()


EventActionMap = typed_string_map(text_type)


class InputEventBindings(JsonObjectWrapper):
    __slots__ = ()
    viewer = wrapped_property('viewer', EventActionMap)
    slice_view = sliceView = wrapped_property('sliceView', EventActionMap)
    perspective_view = perspectiveView = wrapped_property('perspectiveView', EventActionMap)
    data_view = dataView = wrapped_property('dataView', EventActionMap)


class PrefetchState(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    priority = wrapped_property('priority', optional(int, 0))
    state = wrapped_property('state', viewer_state.ViewerState)


class ScaleBarOptions(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    scale_factor = scaleFactor = wrapped_property('scaleFactor', optional(float, 1))
    text_height_in_pixels = textHeightInPixels = wrapped_property('textHeightInPixels',
                                                                  optional(float, 15))
    bar_height_in_pixels = barHeightInPixels = wrapped_property('barHeightInPixels',
                                                                optional(float, 8))
    bar_top_margin_in_pixels = barTopMarginInPixels = wrapped_property(
        'barTopMarginInPixels', optional(float, 5))
    font_name = fontName = wrapped_property('fontName', optional(text_type, 'sans-serif'))
    padding_in_pixels = paddingInPixels = wrapped_property('paddingInPixels', optional(float, 2))
    max_width_in_pixels = maxWidthInPixels = wrapped_property('maxWidthInPixels',
                                                              optional(int, 100))
    max_width_fraction = maxWidthFraction = wrapped_property('maxWidthFraction',
                                                             optional(float, 0.25))
    left_pixel_offset = leftPixelOffset = wrapped_property('leftPixelOffset', optional(int, 10))
    bottom_pixel_offset = bottomPixelOffset = wrapped_property('bottomPixelOffset',
                                                               optional(int, 10))


class ConfigState(JsonObjectWrapper):
    __slots__ = ()
    credentials = wrapped_property('credentials', typed_string_map(dict))
    actions = wrapped_property('actions', typed_set(text_type))
    input_event_bindings = inputEventBindings = wrapped_property('inputEventBindings',
                                                                 InputEventBindings)
    status_messages = statusMessages = wrapped_property('statusMessages',
                                                        typed_string_map(text_type))
    source_generations = sourceGenerations = wrapped_property('sourceGenerations',
                                                              typed_string_map(int))
    screenshot = wrapped_property('screenshot', optional(text_type))
    show_ui_controls = showUIControls = wrapped_property('showUIControls', optional(bool, True))
    show_location = showLocation = wrapped_property('showLocation', optional(bool, True))
    show_layer_panel = showLayerPanel = wrapped_property('showLayerPanel', optional(bool, True))
    show_help_button = showHelpButton = wrapped_property('showHelpButton', optional(bool, True))
    show_settings_button = showSettingsButton = wrapped_property('showSettingsButton', optional(bool, True))
    show_layer_side_panel_button = showLayerSidePanelButton = wrapped_property('showLayerSidePanelButton', optional(bool, True))
    show_layer_list_panel_button = showLayerListPanelButton = wrapped_property('showLayerListPanelButton', optional(bool, True))
    show_selection_panel_button = showSelectionPanelButton = wrapped_property('showSelectionPanelButton', optional(bool, True))
    show_panel_borders = showPanelBorders = wrapped_property('showPanelBorders',
                                                             optional(bool, True))
    scale_bar_options = scaleBarOptions = wrapped_property('scaleBarOptions', ScaleBarOptions)
    show_layer_hover_values = showLayerHoverValues = wrapped_property('showLayerHoverValues',
                                                                      optional(bool, True))
    viewer_size = viewerSize = wrapped_property('viewerSize', optional(array_wrapper(np.int64, 2)))
    prefetch = wrapped_property('prefetch', typed_list(PrefetchState))


class PrivateState(JsonObjectWrapper):
    __slots__ = ()
    credentials = wrapped_property('credentials', typed_string_map(optional(int)))



"""Writes annotations in the Precomputed annotation format.

This provides a simple way to write annotations in the precomputed format, but
has a number of limitations that makes it suitable only for a relatively small
amount of annotation data:

- All annotations are buffered in memory.

- Only a trivial spatial index consisting of a single grid cell at a single
  level is generated.  Consequently, Neuroglancer will be forced to download all
  annotations at once.

- All indices are written in the unsharded format.  Consequently, there is at
  least one file written per annotation.
"""

from . import coordinate_space
from typing import List, Sequence, NamedTuple, Optional
from typing_exceptions import Literal
from . import viewer_state
import numbers
import io
import json
import os
import numpy as np
import struct


class Annotation(NamedTuple):
    id: int
    encoded: bytes
    relationships: Sequence[Sequence[int]]


_PROPERTY_DTYPES = {
    'uint8': (('|u1', ), 1),
    'uint16': (('<u2', ), 2),
    'uint32': (('<u4', ), 3),
    'int8': (('|i1', ), 1),
    'int16': (('<i2', ), 2),
    'int32': (('<i4', ), 4),
    'float32': (('<f4', ), 4),
    'rgb': (('|u1', (3, )), 1),
    'rgba': (('|u1', (4, )), 1),
}

AnnotationType = Literal['point', 'line', 'axis_aligned_bounding_box', 'ellipsoid']


def _get_dtype_for_geometry(annotation_type: AnnotationType, rank: int):
    geometry_size = rank if annotation_type == 'point' else 2 * rank
    return [('geometry', '<f4', geometry_size)]


def _get_dtype_for_properties(properties: Sequence[viewer_state.AnnotationPropertySpec]):
    dtype = []
    offset = 0
    for i, p in enumerate(properties):
        dtype_entry, alignment = _PROPERTY_DTYPES[p.type]
        if offset % alignment:
            padded_offset = (offset + alignment - 1) // alignment * alignment
            padding = padded_offset - offset
            dtype.append((f'padding{offset}', '|u1', (padding, )))
            offset += padding
        dtype.append((f'property{i}', *dtype_entry))
        size = np.dtype(dtype[-1:]).itemsize
        offset += size
    alignment = 4
    if offset % alignment:
        padded_offset = (offset + alignment - 1) // alignment * alignment
        padding = padded_offset - offset
        dtype.append((f'padding{offset}', '|u1', (padding, )))
        offset += padding
    return dtype


class AnnotationWriter:
    def __init__(
            self,
            coordinate_space: coordinate_space.CoordinateSpace,
            annotation_type: AnnotationType,
            relationships: Sequence[str] = (),
            properties: Sequence[viewer_state.AnnotationPropertySpec] = (),
    ):
        self.coordinate_space = coordinate_space
        self.relationships = list(relationships)
        self.annotation_type = annotation_type
        self.properties = list(properties)
        self.properties.sort(key=lambda p: -_PROPERTY_DTYPES[p.type][1])
        self.annotations = []
        self.rank = coordinate_space.rank
        self.dtype = (_get_dtype_for_geometry(annotation_type, coordinate_space.rank) +
                      _get_dtype_for_properties(self.properties))
        self.lower_bound = np.full(shape=(self.rank, ), fill_value=float('inf'), dtype=np.float32)
        self.upper_bound = np.full(shape=(self.rank, ), fill_value=float('-inf'), dtype=np.float32)
        self.related_annotations = [{} for _ in self.relationships]

    def add_point(self, point: Sequence[int], id: Optional[int] = None, **kwargs):
        if self.annotation_type != 'point':
            raise ValueError(
                f'Expected annotation type point, but received: {self.annotation_type}'
            )
        if len(point) != self.coordinate_space.rank:
            raise ValueError(
                f'Expected point to have length {self.coordinate_space.rank}, but received: {len(point)}'
            )

        self.lower_bound = np.minimum(self.lower_bound, point)
        self.upper_bound = np.maximum(self.upper_bound, point)
        self._add_obj(point, id, kwargs)
        
    def add_axis_aligned_bounding_box(self, point_a: Sequence[int], point_b: Sequence[int], id: Optional[int] = None, **kwargs):
        if self.annotation_type != 'axis_aligned_bounding_box':
            raise ValueError(
                f'Expected annotation type axis_aligned_bounding_box, but received: {self.annotation_type}'
            )
        self._add_two_point_obj(point_a, point_b, id, kwargs)
    
    def add_line(self, point_a: Sequence[int], point_b: Sequence[int], id: Optional[int] = None, **kwargs):
        if self.annotation_type != 'line':
            raise ValueError(
                f'Expected annotation type line, but received: {self.annotation_type}'
            )
        self._add_two_point_obj(point_a, point_b, id, kwargs)

    def _add_two_point_obj(self, point_a: Sequence[int], point_b: Sequence[int], id: Optional[int] = None, **kwargs):
        if len(point_a) != self.coordinate_space.rank:
            raise ValueError(
                f'Expected coordinates to have length {self.coordinate_space.rank}, but received: {len(point_a)}'
            )
            
        if len(point_b) != self.coordinate_space.rank:
            raise ValueError(
                f'Expected coordinates to have length {self.coordinate_space.rank}, but received: {len(point_b)}'
            )
            
        self.lower_bound = np.minimum(self.lower_bound, point_a)
        self.upper_bound = np.maximum(self.upper_bound, point_b)
        coords = np.concatenate((point_a, point_b))
        self._add_obj(coords, id, kwargs)

    def _add_obj(self, coords: Sequence[int], id: Optional[int], **kwargs):
        encoded = np.zeros(shape=(), dtype=self.dtype)
        encoded[()]['geometry'] = point

        for i, p in enumerate(self.properties):
            if p.id in kwargs:
                encoded[()][f'property{i}'] = kwargs.pop(p.id)

        related_ids = []
        for relationship in self.relationships:
            ids = kwargs.pop(relationship, None)
            if ids is None:
                ids = []
            if isinstance(ids, numbers.Integral):
                ids = [ids]
            related_ids.append(ids)

        if kwargs:
            raise ValueError(f'Unexpected keyword arguments {kwargs}')

        if id is None:
            id = len(self.annotations)

        annotation = Annotation(id=id, encoded=encoded.tobytes(), relationships=related_ids)

        self.annotations.append(annotation)

        for i, segment_ids in enumerate(related_ids):
            for segment_id in segment_ids:
                rel_index = self.related_annotations[i]
                rel_index.setdefault(segment_id, [])
                rel_index.append(annotation)

    def _serialize_annotations(self, f, annotations: List[Annotation]):
        f.write(struct.pack('<Q', len(annotations)))
        for annotation in annotations:
            f.write(annotation.encoded)
        for annotation in annotations:
            f.write(struct.pack('<Q', annotation.id))

    def _serialize_annotation(self, f, annotation: Annotation):
        f.write(annotation.encoded)
        for related_ids in annotation.relationships:
            f.write(struct.pack('<I', len(related_ids)))
            for related_id in related_ids:
                f.write(struct.pack('<Q', related_id))

    def write(self, path: str):

        metadata = {
            '@type':
            'neuroglancer_annotations_v1',
            'dimensions':
            self.coordinate_space.to_json(),
            'lower_bound': [float(x) for x in self.lower_bound],
            'upper_bound': [float(x) for x in self.upper_bound],
            'annotation_type':
            self.annotation_type,
            'properties': [p.to_json() for p in self.properties],
            'relationships': [{
                'id': relationship,
                'key': f'rel_{relationship}'
            } for relationship in self.relationships],
            'by_id': {
                'key': 'by_id',
            },
            'spatial': [
                {
                    'key': 'spatial0',
                    'grid_shape': [1] * self.rank,
                    'chunk_size': [max(1, float(x)) for x in self.upper_bound - self.lower_bound],
                    'limit': len(self.annotations),
                },
            ],
        }

        os.makedirs(path, exist_ok=True)
        for relationship in self.relationships:
            os.makedirs(os.path.join(path, f'rel_{relationship}'), exist_ok=True)
        os.makedirs(os.path.join(path, 'by_id'), exist_ok=True)
        os.makedirs(os.path.join(path, 'spatial0'), exist_ok=True)

        with open(os.path.join(path, 'info'), 'w') as f:
            f.write(json.dumps(metadata))

        with open(os.path.join(path, 'spatial0', '_'.join('0' for _ in range(self.rank))),
                  'wb') as f:
            self._serialize_annotations(f, self.annotations)

        for annotation in self.annotations:
            with open(os.path.join(path, 'by_id', str(annotation.id)), 'wb') as f:
                self._serialize_annotation(f, annotation)

        for i, relationship in enumerate(self.relationships):
            rel_index = self.related_annotations[i]
            for segment_id, annotations in rel_index.items():
                with open(os.path.join(path, f'rel_{relationship}', str(segment_id)), 'wb') as f:
                    self._serialize_annotations(f, annotations)
